@@ -26,7 +26,15 @@
 
 #if HAVE_CONFIG_H
 # include "config.h"
-#endif /* HAVE_CONFIG_H */
+#endif
+
+#include <assert.h>
+#include <errno.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+
+#include <time.h>
 
 #include "aes.h"
 #include "arcfour.h"
@@ -41,40 +49,50 @@
 
 #include "cbc.h"
 
-#include <assert.h>
-#include <errno.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
 
-#include <time.h>
-
-#if __GNUC__
-# define UNUSED __attribute__ ((__unused__))
-#else
-# define UNUSED
-#endif
-
-/* Encrypt 10MB, 1K at a time. */
-#define BENCH_BLOCK 1024
-#define BENCH_COUNT 10240
+/* Process BENCH_BLOCK bytes at a time, for BENCH_INTERVAL clocks. */
+#define BENCH_BLOCK 10240
+#define BENCH_INTERVAL (CLOCKS_PER_SEC / 4)
 
 /* Total MB:s, for MB/s figures. */
 #define BENCH_TOTAL 10.0
 
+/* Returns second per function call */
 static double
 time_function(void (*f)(void *arg), void *arg)
 {
   clock_t before;
   clock_t after;
-
+  clock_t done;
+  unsigned ncalls;
+  
   before = clock();
+  done = before + BENCH_INTERVAL;
+  ncalls = 0;
   
-  f(arg);
+  do 
+    {
+      f(arg);
+      after = clock();
+      ncalls++;
+    }
+  while (after < done);
   
-  after = clock();
+  return ((double)(after - before)) / CLOCKS_PER_SEC / ncalls;
+}
 
-  return ((double)(after - before)) / CLOCKS_PER_SEC;
+struct bench_hash_info
+{
+  void *ctx;
+  nettle_hash_update_func update;
+  const uint8_t *data;
+};
+
+static void
+bench_hash(void *arg)
+{
+  struct bench_hash_info *info = arg;
+  info->update(info->ctx, BENCH_BLOCK, info->data);
 }
 
 struct bench_cipher_info
@@ -88,10 +106,7 @@ static void
 bench_cipher(void *arg)
 {
   struct bench_cipher_info *info = arg;
-  unsigned i;
-  
-  for (i = 0; i<BENCH_COUNT; i++)
-    info->crypt(info->ctx, BENCH_BLOCK, info->data, info->data);
+  info->crypt(info->ctx, BENCH_BLOCK, info->data, info->data);
 }
 
 struct bench_cbc_info
@@ -109,24 +124,18 @@ static void
 bench_cbc_encrypt(void *arg)
 {
   struct bench_cbc_info *info = arg;
-  unsigned i;
-
-  for (i = 0; i<BENCH_COUNT; i++)
-    cbc_encrypt(info->ctx, info->crypt,
-		info->block_size, info->iv,
-		BENCH_BLOCK, info->data, info->data);
+  cbc_encrypt(info->ctx, info->crypt,
+	      info->block_size, info->iv,
+	      BENCH_BLOCK, info->data, info->data);
 }
 
 static void
 bench_cbc_decrypt(void *arg)
 {
   struct bench_cbc_info *info = arg;
-  unsigned i;
-
-  for (i = 0; i<BENCH_COUNT; i++)
-    cbc_decrypt(info->ctx, info->crypt,
-		info->block_size, info->iv,
-		BENCH_BLOCK, info->data, info->data);
+  cbc_decrypt(info->ctx, info->crypt,
+	      info->block_size, info->iv,
+	      BENCH_BLOCK, info->data, info->data);
 }
 
 /* Set data[i] = floor(sqrt(i)) */
@@ -153,20 +162,51 @@ init_key(unsigned length,
 
 static void
 display(const char *name, const char *mode,
-	double elapsed)
+	double speed)
 {
-  printf("%13s (%s): %.2fs, %.3fMB/s\n",
+  printf("%13s (%s): %.3fMB/s\n",
 	 name, mode,
-	 elapsed, BENCH_TOTAL / elapsed);
+	 1 / (speed * 1048576.0 / BENCH_BLOCK));
+}
+
+static void *
+xalloc(size_t size)
+{
+  void *p = malloc(size);
+  if (!p)
+    {
+      fprintf(stderr, "Virtual memory exhausted.\n");
+      abort();
+    }
+
+  return p;
+}
+
+static void
+time_hash(const struct nettle_hash *hash)
+{
+  static uint8_t data[BENCH_BLOCK];
+  struct bench_hash_info info;
+  info.ctx = xalloc(hash->context_size); 
+  info.update = hash->update;
+  info.data = data;
+
+  init_data(data);
+  hash->init(info.ctx);
+
+  display(hash->name, "Update",
+	  time_function(bench_hash, &info));
+
+  free(info.ctx);
 }
 
 static void
 time_cipher(const struct nettle_cipher *cipher)
 {
-  void *ctx = alloca(cipher->context_size);
-  uint8_t *key = alloca(cipher->key_size);
+  void *ctx = xalloc(cipher->context_size);
+  uint8_t *key = xalloc(cipher->key_size);
 
-  uint8_t data[BENCH_BLOCK];
+  static uint8_t data[BENCH_BLOCK];
 
   printf("\n");
   
@@ -201,7 +241,7 @@ time_cipher(const struct nettle_cipher *cipher)
 
   if (cipher->block_size)
     {
-      uint8_t *iv = alloca(cipher->block_size);
+      uint8_t *iv = xalloc(cipher->block_size);
       
       /* Do CBC mode */
       {
@@ -235,7 +275,10 @@ time_cipher(const struct nettle_cipher *cipher)
 	display(cipher->name, "CBC decrypt",
 		time_function(bench_cbc_decrypt, &info));
       }
+      free(iv);
     }
+  free(ctx);
+  free(key);
 }
 
 #if HAVE_LIBCRYPTO
@@ -248,6 +291,14 @@ int
 main(int argc UNUSED, char **argv UNUSED)
 {
   unsigned i;
+
+  const struct nettle_hash *hashes[] =
+    {
+      &nettle_md2, &nettle_md4, &nettle_md5,
+      &nettle_sha1, &nettle_sha256,
+      NULL
+    };
+
   const struct nettle_cipher *ciphers[] =
     {
       &nettle_aes128, &nettle_aes192, &nettle_aes256,
@@ -262,6 +313,9 @@ main(int argc UNUSED, char **argv UNUSED)
       NULL
     };
 
+  for (i = 0; hashes[i]; i++)
+    time_hash(hashes[i]);
+  
   for (i = 0; ciphers[i]; i++)
     time_cipher(ciphers[i]);
   
